/*
 *  Ray Trace Mandlebulb using CPU, OpenMP or CUDA
 *  Many functions run on BOTH the CPU and GPU
 *
 *  'm' switches between CPU, OpenMP and CUDA
 *  '0' snaps angles to 0,0
 *  arrows to rotate the world
 *  wasd   move light
 *  +/-    adjust power
 *  []     adjust iterations
 */

using namespace std;

#include "CSCIx239.h"
#include <vector>
#include <cuda.h>
#include "Ray.h"

//  Global constants
#define Rmax    2
#define MaxDist 4
#define MaxStep 100

//  Struct with values needed for mandelbulb
class MandelBulb
{
   public:
      int   n;
      int   maxiter;
      bool  fast;
      bool  shadow;
      Vec3  L;
      Mat3  rot;
      float zoom;
      int   wid,hgt,dim;
};

//  Global variables
int mode=2;                 //  CPU or GPU
int th=0;                   //  Azimuth of view angle
int ph=0;                   //  Elevation of view angle
int Th=30;                  //  Azimuth of light angle
int Ph=30;                  //  Elevation of light angle
unsigned char* pixels=NULL; //  Pixel array
MandelBulb mb;              //  Mandelbulb structure
int nthread;                //  Threads per block
char* text[] = {"CPU","OpenMP","CUDA"};

/*
 *  Initialize fastest GPU device
 */
int InitGPU(int verbose)
{
   //  Get number of CUDA devices
   int num;
   if (cudaGetDeviceCount(&num)) Fatal("Cannot get number of CUDA devices\n");
   if (num<1) Fatal("No CUDA devices found\n");

   //  Get fastest device
   cudaDeviceProp prop;
   int   MaxDevice = -1;
   int   MaxGflops = -1;
   for (int dev=0;dev<num;dev++)
   {
      if (cudaGetDeviceProperties(&prop,dev)) Fatal("Error getting device %d properties\n",dev);
      int Gflops = prop.multiProcessorCount * prop.clockRate;
      if (verbose) printf("CUDA Device %d: %s Gflops %f Processors %d Threads/Block %d\n",dev,prop.name,1e-6*Gflops,prop.multiProcessorCount,prop.maxThreadsPerBlock);
      if(Gflops > MaxGflops)
      {
         MaxGflops = Gflops;
         MaxDevice = dev;
      }
   }

   //  Print and set device
   if (cudaGetDeviceProperties(&prop,MaxDevice)) Fatal("Error getting device %d properties\n",MaxDevice);
   printf("Fastest CUDA Device %d: %s\n",MaxDevice,prop.name);
   cudaSetDevice(MaxDevice);

   //  Return max thread count
   return prop.maxThreadsPerBlock;
}

/*
 *  Mandlebulb distance estimate
 */
__host__ __device__ inline float Step(const Vec3& C,const MandelBulb& mb)
{
   //  Skip one iteration (Z = 0 + C)
   Vec3  Z  = C;
   //  Compute |Z|
   float R  = sqrt(Z.x*Z.x+Z.y*Z.y+Z.z*Z.z);
   //  Initialize |Z'|
   float dR = 1;
   //  Loop maxiter times or until divergence
   int   i = mb.maxiter;
   while (R<Rmax && i--)
   {
      //  Convert to spherical
      float ph = asinf(Z.z/R);
      float th = atan2f(Z.y,Z.x);
      //  Compute derivative recursively n|Z^(n-1)|Z'
      float Rn = powf(R,mb.n-1);
      dR = Rn*dR*mb.n + 1;
      //  Compute R^n
      Rn *= R;
      //  Compute sin(n*ph) & cos(n*ph)
      float sinph,cosph;
      sincosf(mb.n*ph,&sinph,&cosph);
      //  Compute sin(n*th) & cos(n*th)
      float sinth,costh;
      sincosf(mb.n*th,&sinth,&costh);
      //  Compute Z^n+C
      Z = Rn*Vec3(cosph*costh,cosph*sinth,sinph) + C;
      //  Compute |Z|
      R = sqrt(Z.x*Z.x+Z.y*Z.y+Z.z*Z.z);
   }
   //  Evaluate G(c)/G'(c)
   return 0.5f*log(R)*R/dR;
}

/*
 *  Mandelbulb normal vector
 */
__host__ __device__ inline Vec3 Normal(const Vec3 &C,const MandelBulb& mb)
{
   const float d = 1e-3;
   if (mb.fast)
   {
      float c  = Step(C,mb);
      float cx = Step(C+Vec3(d,0,0),mb);
      float cy = Step(C+Vec3(0,d,0),mb);
      float cz = Step(C+Vec3(0,0,d),mb);
      return normalize(Vec3(c-cx,c-cy,c-cz));
   }
   else
   {
      float cx = Step(C-Vec3(d,0,0),mb) - Step(C+Vec3(d,0,0),mb);
      float cy = Step(C-Vec3(0,d,0),mb) - Step(C+Vec3(0,d,0),mb);
      float cz = Step(C-Vec3(0,0,d),mb) - Step(C+Vec3(0,0,d),mb);
      return normalize(Vec3(cx,cy,cz));
   }
}

/*
 *  Adaptive ray marching
 */
__host__ __device__ inline bool RayMarchA(const Vec3& org, const Vec3& dir,const MandelBulb& mb,Vec3& point)
{
   const float eps = 1e-3;
   float dist  = 0;
   point = org;
   int k=MaxStep;
   //  Estimate distance until max distance or max steps
   while (dist<MaxDist && k--)
   {
      float step = Step(point,mb);
      point += step*dir;
      dist  += step;
      //  Stop of step is small
      if (fabs(step)<eps) return true;
   }
   return false;
}

/*
 *  Non-adaptive ray marching
 */
__host__ __device__ inline bool RayMarchNA(Vec3 point,const Vec3& dir,const MandelBulb& mb)
{
   if (!mb.shadow) return false;
   float step = 0.1;
   float dist = 0;
   const float eps = 0.1*step;
   int k=MaxStep;
   //  Estimate distance until max distance or max steps
   while (dist<MaxDist && k--)
   {
      point += step*dir;
      dist  += step;
      step *= 1.1f; //increase step regularly (ie resolution decrease with distance)
      if (Step(point,mb)<eps) return true;
   }
   return false;
}

/*
 *  Ray trace with no secondary rays
 */
__host__ __device__ Color RayTrace(Vec3 org,Vec3 dir,const MandelBulb& mb)
{
   //  Scalar intensity
   float f=0;
   Vec3  P;
   //  Test for intersection
   if (RayMarchA(org,dir,mb,P))
   {
      //  Add ambient
      f = 0.3;
      //  Compute normal vector
      Vec3 N = Normal(P,mb);
      //  Compute cosine with light vector
      float I = -(mb.L*N);
      //  If illuminated and not in shadow, add diffuse
      if (I>0 && !RayMarchNA(P,mb.L,mb)) f += 0.7*I;
   }
   //  Return the resulting color
   return Color(f,f,0);
}

/*
 *  Ray trace a pixel
 */
__host__ __device__ void RayTracePixel(unsigned char* pix,int k,const MandelBulb& mb)
{
   //  Translate scalar index into pixel
   int i = k%mb.wid;
   int j = k/mb.wid;
   // Cast the intial ray from (x,y,1.1) down z axis
   Vec3 org(mb.zoom*(i-mb.wid/2) , mb.zoom*(j-mb.hgt/2) , 1.1);
   Vec3 dir(0,0,-1);
   Color col = RayTrace(mb.rot*org , mb.rot*dir , mb);
   //  Copy color to pixel array
   pix += 4*k;
   *pix++ = col.r>1 ? 255 : (unsigned int)(255*col.r);
   *pix++ = col.g>1 ? 255 : (unsigned int)(255*col.g);
   *pix++ = col.b>1 ? 255 : (unsigned int)(255*col.b);
   *pix++ = 255;
}

/*
 *  Set rotation matrix
 */
void SetRot(void)
{
   double M[16];
   //  Initialize Modelview
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   //  Apply rotations
   glRotated(-ph,1,0,0);
   glRotated(th,0,1,0);
   //  Copy matrix to row vectors
   glGetDoublev(GL_MODELVIEW_MATRIX,M);
   mb.rot.x.x = M[0]; mb.rot.x.y = M[4]; mb.rot.x.z = M[8];
   mb.rot.y.x = M[1]; mb.rot.y.y = M[5]; mb.rot.y.z = M[9];
   mb.rot.z.x = M[2]; mb.rot.z.y = M[6]; mb.rot.z.z = M[10];
   //  Reset
   glLoadIdentity();
}

/*
 *  GLUT calls this routine when an arrow key is pressed
 */
void special(int key,int x,int y)
{
   //  Right arrow key - increase angle by 5 degrees
   if (key == GLUT_KEY_RIGHT)
      th += 5;
   //  Left arrow key - decrease angle by 5 degrees
   else if (key == GLUT_KEY_LEFT)
      th -= 5;
   //  Up arrow key - increase elevation by 5 degrees
   else if (key == GLUT_KEY_UP)
      ph += 5;
   //  Down arrow key - decrease elevation by 5 degrees
   else if (key == GLUT_KEY_DOWN)
      ph -= 5;
   //  Page Up key - increase zoom
   else if (key == GLUT_KEY_PAGE_UP)
      mb.zoom *= 0.9;
   //  Page Down key - decrease zoom
   else if (key == GLUT_KEY_PAGE_DOWN)
      mb.zoom *= 1.1;
   //  Keep angles to +/-360 degrees
   th %= 360;
   ph %= 360;
   SetRot();
   //  Tell GLUT it is necessary to redisplay the scene
   glutPostRedisplay();
}

/*
 *  GLUT calls this routine when a key is pressed
 */
void key(unsigned char ch,int x,int y)
{
   //  Exit on ESC
   if (ch == 27)
      exit(0);
   //  Reset view angle
   else if (ch == '0')
      th = ph = 0;
   //  CPU/GPU
   else if (ch == 'm')
      mode = (mode+1)%3;
   else if (ch == 'M')
      mode = (mode+2)%3;
   //  Change power
   else if (ch == '-' && mb.n>1)
      mb.n--;
   else if (ch == '+')
      mb.n++;
   //  Change maximum iterations
   else if (ch == '[' && mb.maxiter>0)
      mb.maxiter--;
   else if (ch == ']')
      mb.maxiter++;
   //  Increase angle by 5 degrees
   else if (ch == 'a')
      Th += 5;
   //  Decrease angle by 5 degrees
   else if (ch == 'd')
      Th -= 5;
   //  Increase elevation by 5 degrees
   else if (ch == 'w')
      Ph += 5;
   //  Decrease elevation by 5 degrees
   else if (ch == 's')
      Ph -= 5;
   //  Decrease elevation by 5 degrees
   else if (ch == ' ')
      mb.shadow = !mb.shadow;
   else if (ch == '.')
      mb.fast = !mb.fast;
   //  Reset Angle
   SetRot();
   //  Tell GLUT it is necessary to redisplay the scene
   glutPostRedisplay();
}

/*
 *  GLUT calls this routine when the window is resized
 */
void reshape(int width,int height)
{
   //  Window dimensions
   mb.wid = width;
   mb.hgt = height;
   mb.dim = 4*width*height;
   //  Allocate pixels size of window
   delete pixels;
   pixels = new unsigned char [mb.dim];
   //  Set the viewport to the entire window
   glViewport(0,0, width,height);
}

/*
 *  Initialize
 */
void Init(MandelBulb& mb)
{
   //  Mandelbulb parameters
   mb.n=6;             //  Power
   mb.maxiter=10;      //  Max iterations
   mb.shadow=true;     //  Shadows
   mb.fast=true;       //  Fast normals
   mb.zoom=0.005;      //  Zoom level
   //  Projection
   SetRot();
}

/*
 *  Ray trace the scene host/device interface
 */
__global__ void RayTracePixDev(unsigned char* pix,const MandelBulb mb)
{
   unsigned int k = blockIdx.x*blockDim.x+threadIdx.x;
   if (k<mb.wid*mb.hgt)
      RayTracePixel(pix,k,mb);
}

/*
 *  OpenGL (GLUT) calls this routine to display the scene
 */
void display()
{
   Elapsed();
   //  Set light direction
   mb.L.x = Sin(Th)*Cos(Ph);
   mb.L.y =         Sin(Ph);
   mb.L.z = Cos(Th)*Cos(Ph);

   //  CPU
   if (mode==0)
   {
      for (int k=0;k<mb.hgt*mb.wid;k++)
         RayTracePixel(pixels,k,mb);
   }
   //  OpenMP
   else if (mode==1)
   {
      #pragma omp parallel for
      for (int k=0;k<mb.hgt*mb.wid;k++)
         RayTracePixel(pixels,k,mb);
   }
   //  CUDA
   else
   {
      //  Allocate pixel array on device
      unsigned char* devpix;
      if (cudaMalloc((void**)&devpix,mb.dim)) Fatal("Cannot allocate device memory for pixels\n");

      //  Ray trace scene
      int nblock = mb.wid*mb.hgt/nthread;
      if (nblock*nthread<mb.wid*mb.hgt) nblock++;
      RayTracePixDev<<<nblock,nthread>>>(devpix,mb);

      //  Copy pixels from device to host
      if (cudaMemcpy(pixels,devpix,mb.dim,cudaMemcpyDeviceToHost)) Fatal("Cannot copy pixels from device to host\n");

      //  Free device memory
      cudaFree(devpix);
   }

   //  Time ray tracing
   double t = Elapsed();
   //  Blit scene to screen
   glWindowPos2i(0,0);
   glDrawPixels(mb.wid,mb.hgt,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
   //  Display
   glWindowPos2i(5,5);
   Print("Size %dx%d Time %.3fs Angle %d,%d Power %d MaxIter %d Mode %s",mb.wid,mb.hgt,t,th,ph,mb.n,mb.maxiter,text[mode]);
   //  Flush
   glFlush();
   glutSwapBuffers();
}

/*
 *  Start up GLUT and tell it what to do
 */
int main(int argc,char* argv[])
{
   //  Initialize GLUT and process user parameters
   glutInit(&argc,argv);
   //  Request double buffered, true Color window with Z buffering at 600x600
   glutInitWindowSize(800,600);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
   //  Create the window
   glutCreateWindow("Mandelbulb");
   //  Set GLUT callbacks
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutSpecialFunc(special);
   glutKeyboardFunc(key);
   //  Initialize Mandelbulb
   Init(mb);
   //  Initialize CUDA
   nthread = InitGPU(1)/2;
   //  Pass control to GLUT so it can interact with the user
   glutMainLoop();
   return 0;
}
